# Creating an AI Tools

## Mock Example
In this example, we are creating a simple AI Tool for querying JIRA. When the tool is called, simply print the JQL query generated by OpenAI. But you can easily replace the JIRATool.invoke implementation with real JIA Query.

```python
%python%

from pydantic import BaseModel, Field, ConfigDict
from webcli2.core.ai import AITool, AIAgent

# You need to define the input type your tool
class JQLInput(BaseModel):
    jql_query: str = Field(..., description="The JQL Query String") # Make the description easy to understand by OpenAI
    model_config = ConfigDict(extra='forbid')  # This is required by OpenAI


class JIRATool(AITool):
    def __init__(self, *, name:str):
        super().__init__(
            name=name, 
            description="Execute JQL Query",  # Make sure the description is understood by OpenAI, so it knows to call this tool
                                              # to execute JQL Query
            input_type=JQLInput
        )

    def invoke(self, input:JQLInput):
        print(f"running query: {input.jql_query}")

ai_agent = create_ai_agent()  # create an AI Agent
run_jql_query = JIRATool(name="run_jql_query") # Create a AI Tool for executeing JQL query
ai_agent.register_tool(run_jql_query) # Register this AI tool with AI Agent

question = """
Can you generate a JQL query, the query need to return all JIRA ticket for project HWD, and assignee is shizhong, let's sort the result by created field, with most recent created on top. Use the provided tool to fetch data.
"""
r = ai_agent.ask_ai(question) # Now ask OpenAI
ai_agent.process_tools_response(r) # Now process the OpenAI response, which will invoke AI Tool if instructed by OpenAI
```

## Real Example
In this example, instead of mocking the JQL query, we are going to execute the JQL query.

```python
%python%

# Make sure you have jira installed: pip install jira

from pydantic import BaseModel, Field, ConfigDict
from webcli2.core.ai import AITool, AIAgent
from jira import JIRA

# You need to define the input of your tool
class JQLInput(BaseModel):
    jql_query: str = Field(..., description="The JQL Query String")
    model_config = ConfigDict(extra='forbid')


class JIRATool(AITool):
    jira_client:JIRA

    def __init__(self, *, name:str):
        super().__init__(name=name, description="Execute JQL Query", input_type=JQLInput)
        options = {"server": "https://jira.oci.oraclecorp.com"}
        API_TOKEN = "***"
        self.jira_client = JIRA(options, token_auth=API_TOKEN)

    def invoke(self, input:JQLInput):
        issues = self.jira_client.search_issues(input.jql_query, maxResults=100)
        print("JIRA Issues:")
        print("------------------------")
        for issue in issues:
            print(f"{issue.key} | {issue.fields.summary}")
        print("------------------------")


ai_agent = create_ai_agent()
run_jql_query = JIRATool(name="run_jql_query")
ai_agent.register_tool(run_jql_query)

question = """
Can you generate a JQL query, the query need to return all JIRA ticket for project HWD, and assignee is shizhong, let's sort the result by created field, with most recent created on top. Use the provided tool to fetch data.
"""
r = ai_agent.ask_ai(question)
ai_agent.process_tools_response(r)
```

Here is the screenshot when you run this AI Agent:
![Screenshot](ai-agent-01.png "Screenshot")

